/* wptClipSignDlg.cpp - WinPT clipboard sign dialog
 *	Copyright (C) 2000-2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptAgent.h"
#include "wptNLS.h"
#include "wptCrypto.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptRegistry.h"
#include "wptKeylist.h"
#include "wptErrors.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptContext.h" /* for passphrase_s */


/* Sign the clipboard contents with the key @keyid and wrap
   text lines to @wraplen (0 disable line wrapping).
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_sign (gpgme_ctx_t ctx, const char *keyid, int wraplen)
{
    gpgme_error_t err;
    gpgme_data_t plain = NULL;
    gpgme_data_t sig = NULL;
    gpgme_key_t key = NULL;
    
    if (!keyid)
        return gpg_error (GPG_ERR_INV_ARG);
    
    gpgme_set_armor (ctx, 1);
    err = gpg_data_utf8_new_from_clipboard (&plain, wraplen, NULL);
    if (err)
        return err;
    err = get_pubkey (keyid, &key);
    if (err)
	goto leave;
    err = gpgme_signers_add (ctx, key);
    if (err)
	goto leave;
    err = gpgme_data_new (&sig);
    if (err)
	goto leave;
    err = gpgme_op_sign (ctx, plain, sig, GPGME_SIG_MODE_CLEAR);
    if (err)
	goto leave;
    
    gpg_data_release_to_clipboard (sig, 1);
    sig = NULL;    

leave:
    if (plain)
	gpgme_data_release (plain);
    if (sig)
	gpgme_data_release (sig);
    return err;
}


/* This function is used when only one secret key is available.
   it doesn't make sense to offer a dialog for this case. */
void
one_key_proc (HWND dlg)
{    
    gpgme_ctx_t ctx;
    gpgme_error_t err;
    passphrase_cb_s pwd;
    char *signer;
    int n = reg_prefs.word_wrap;
    
    signer = get_gnupg_default_key ();
    if (!signer) {
        msg_box (dlg, _("Could not get default key."), _("Signing"), MB_ERR);
        return;
    }
    
    err = gpgme_new (&ctx);
    if (err)
	BUG (dlg);

    set_gpg_passphrase_cb (&pwd, ctx, GPG_CMD_SIGN, dlg, _("Signing"));
    err = gpg_clip_sign (ctx, signer, n);
    if (pwd.cancel)
	goto leave;

    if (gpgme_err_code (err) == GPG_ERR_BAD_PASSPHRASE)
	agent_del_cache (pwd.keyid);
    if (err)
	msg_box (dlg, gpgme_strerror (err), _("Signing"), MB_ERR);
    else
	show_msg (dlg, 1500, _("GnuPG Status: Finished"));
leave:
    gpgme_release (ctx);
    free_if_alloc (signer);
    release_gpg_passphrase_cb (&pwd);
}


/* Count only useable secret keys. 
   Ignore expired, revoked and disabled keys.
   Return value: amount of keys. */
static DWORD
count_useable_seckeys (gpg_keycache_t kc)
{
    struct keycache_s *c;
    DWORD n=0;

    for (c = kc->item; c; c=c->next) {
	if (c->pubpart && key_is_useable (c->pubpart->key))
	    n++;
    }
    return n;
}


static listview_ctrl_t
on_init_dialog (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    listview_ctrl_t lv;
    gpg_keycache_t kc, sec_kc;
    int nkeys;

    SetWindowText (dlg, _("Signing"));
    kc = keycache_get_ctx (KEYCACHE_PUB);
    sec_kc = keycache_get_ctx (KEYCACHE_PRV);

    nkeys = count_useable_seckeys (sec_kc);
    if (nkeys < 1) {
	msg_box (dlg, _("No useable signing key found"), _("Signing"), MB_ERR);
	return NULL;
    }
    else if (nkeys == 1) {
        one_key_proc (dlg);
	return NULL;
    }
    lv = keylist_load (GetDlgItem (dlg, IDC_SIGN_KEYLIST), kc, sec_kc,
			KEYLIST_SIGN, KEY_SORT_USERID);
    center_window (dlg, NULL);
    SetForegroundWindow (dlg);
    return lv;
}


static const char*
lookup_key_userid (const char *patt)
{
    winpt_key_s key;

    memset (&key, 0, sizeof (key));
    if (winpt_get_pubkey (patt, &key))
	return patt;
    return key.ext->uids->uid;
}


/* Dialog box procedure for clipboard signing. */
BOOL CALLBACK 
clip_sign_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{	
    static listview_ctrl_t lv = NULL;    
    gpgme_ctx_t ctx;
    gpgme_error_t err;
    passphrase_cb_s pwd;
    int lv_idx = 0;
    int rc = 0, sel_signer = 0;
    char *signer = NULL;
    
    switch (msg) {
    case WM_INITDIALOG:
	lv = on_init_dialog (dlg, msg, wparam, lparam);
	if (!lv)
	    EndDialog (dlg, TRUE);
        return FALSE;	
        
    case WM_DESTROY:
        if (lv) {
            keylist_delete (lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_NOTIFY:
        NMHDR *notify;
        notify = (NMHDR *)lparam;
        if (notify && notify->code == NM_DBLCLK
            && notify->idFrom == IDC_SIGN_KEYLIST)
            PostMessage (dlg, WM_COMMAND, MAKEWPARAM (IDOK, 0), 0);
        return TRUE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDOK:
	    sel_signer = 1;
            signer = get_gnupg_default_key ();
            if (!signer) {
                msg_box (dlg, _("Could not get default key."), 
			 _("Signing"), MB_ERR);
                return FALSE;
            }
            else if ((lv_idx = listview_get_selected_item (lv)) == -1) {
		rc = log_box (_("Signing"), MB_YESNO,
		              _("No key was chosen.\n"
			        "Use the GPG default key '%s'?"),
			      lookup_key_userid (signer));
                if (rc == IDNO) {
		    free_if_alloc (signer);
                    return FALSE;
		}
                sel_signer = 0;
            }
            if (sel_signer) {
		free_if_alloc (signer);
		signer = new char[32+1];
		if (!signer)
		    BUG (NULL);
                listview_get_item_text (lv, lv_idx, KM_COL_KEYID, signer, 32);
	    }
            err = gpgme_new (&ctx);
	    if (err)
		BUG (NULL);
	    set_gpg_passphrase_cb (&pwd, ctx, GPG_CMD_SIGN, dlg, _("Signing"));
            err = gpg_clip_sign (ctx, signer, reg_prefs.word_wrap);
	    free_if_alloc (signer);
	    release_gpg_passphrase_cb (&pwd);

            if (pwd.cancel && gpgme_err_code (err) == GPG_ERR_BAD_PASSPHRASE) {
		/* The user hit the cancel button or bad passphrase */
		gpgme_release (ctx);
                return TRUE;
            }
            if (err) {
		msg_box (dlg, gpgme_strerror (err), _("Signing"), MB_ERR);
		gpgme_release (ctx);
                return TRUE;
            }
	    else
		show_msg (dlg, 1500, _("GnuPG Status: Finished"));
	    gpgme_release (ctx);
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            EndDialog (dlg, FALSE);
            return FALSE;
        }
        break;
    }
    
    return FALSE;
}
