/* wptClipboard.cpp - GPG related clipboard functions
 *	Copyright (C) 2001, 2002, 2003, 2005, 2006, 2009 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <gpgme.h>

#include "wptW32API.h"
#include "wptCrypto.h"
#include "wptVersion.h"
#include "wptGPG.h"
#include "wptTypes.h"
#include "wptUTF8.h"

/* Default buffer size. */
#define BUFSIZE 192

gpgme_error_t gpg_data_wrap_lines (gpgme_data_t *r_dh, size_t wraplen);

/* XXX: define clipboard errors. */

/* Check if the clipboard contains text. @r_val contains
   one text data is available.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_istext_avail (int *r_val)
{    
    HANDLE clipmem;
    char *clipdata;
    
    if (!r_val)
	return gpg_error (GPG_ERR_INV_ARG);
    *r_val = 0;
    if (OpenClipboard (NULL) == FALSE)
	return gpg_error (GPG_ERR_INTERNAL);
    
    clipmem = GetClipboardData (CF_TEXT);
    if (!clipmem) {
	CloseClipboard();
	return gpg_error (GPG_ERR_NO_DATA);
    }

    clipdata = (char *) GlobalLock (clipmem);
    if (!clipdata) {
	CloseClipboard ();
	return gpg_error (GPG_ERR_INTERNAL);
    }
    
    if (strlen (clipdata) > 0)
	*r_val = 1;
    GlobalUnlock (clipmem);
    CloseClipboard ();
    return 0;
}


/* Figure out what kind of PGP data is stored in @data.
   @r_type contains the ORed types on success.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_parse_pgpid (const char *data, int *r_type)
{
    int type;

    if (!data)
	return gpg_error (GPG_ERR_INV_ARG);
    if (strlen (data) < 19) {
	*r_type = 0;
	return 0;
    }
    type = 0;
    if (strstr (data, "BEGIN PGP MESSAGE"))
	type |= PGP_MESSAGE;
    if (strstr (data, "BEGIN PGP SIGNED MESSAGE"))
	type |= PGP_CLEARSIG;
    if (strstr (data, "BEGIN PGP SIGNATURE"))
	type |= PGP_SIG;
    if (strstr (data, "BEGIN PGP PUBLIC KEY BLOCK"))
	type |= PGP_PUBKEY;
    if (strstr (data, "BEGIN PGP PRIVATE KEY BLOCK") ||
	strstr (data, "BEGIN PGP SECRET KEY BLOCK"))
	type |= PGP_SECKEY;
    if (strstr (data, "- -----BEGIN"))
	type |= PGP_DASH_ESCAPED;
    *r_type = type;
    return 0;
}


/* Check if the clipboard is already OpenPGP protected.
   @r_type optional returns the type and @r_val contains
   1 for yes.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_is_secured (int *r_type, int *r_val)
{
    HANDLE clipmem;
    char *clipdata;
    
    if (!r_val)
	return gpg_error (GPG_ERR_INV_ARG);
    if (OpenClipboard (NULL) == FALSE)
	return gpg_error (GPG_ERR_INTERNAL);
    clipmem = GetClipboardData (CF_TEXT);
    if (!clipmem) {
	*r_val = 0;
	CloseClipboard();
	return gpg_error (GPG_ERR_NO_DATA);
    }
    clipdata = (char *) GlobalLock (clipmem);
    if (!clipdata) {
	*r_val = 0;
	CloseClipboard();
	return gpg_error (GPG_ERR_INTERNAL);
    }
    if (strstr (clipdata, "-----BEGIN PGP") 
	&& strstr (clipdata, "-----END PGP"))
	*r_val = 1;
    GlobalUnlock (clipmem);
    gpg_clip_parse_pgpid (clipdata, r_type);
    CloseClipboard ();
    return gpg_error (GPG_ERR_NO_ERROR);
}


/* Return the OpenPGP data format of the clipboard in @r_type.
   Return 0 on success. */
gpgme_error_t
gpg_clip_get_pgptype (int *r_type)
{
    HANDLE clipmem;
    char *clipdata;
    
    if (OpenClipboard (NULL) == FALSE)
	return gpg_error (GPG_ERR_INTERNAL);
    clipmem = GetClipboardData (CF_TEXT);
    if (!clipmem) {
	CloseClipboard ();
	return gpg_error (GPG_ERR_NO_DATA);
    }
    clipdata = (char *)GlobalLock (clipmem);
    if (!clipdata) {
	CloseClipboard ();
	return gpg_error (GPG_ERR_INTERNAL);
    }    
    gpg_clip_parse_pgpid (clipdata, r_type);
    GlobalUnlock (clipmem);
    CloseClipboard ();
    return gpg_error (GPG_ERR_NO_ERROR);
}


/* Retrieve the text of the clipboard.
   Return value: the text as a string, NULL otherwise. */
static char*
get_w32_clip_text (void)
{
    char *private_clip;
    char *clip;
    size_t size;
    HANDLE cb;
    
    if (!OpenClipboard (NULL))
        return NULL;
    cb = GetClipboardData (CF_TEXT);
    if (!cb) {
	CloseClipboard ();
	return NULL;
    }
    private_clip = (char *)GlobalLock (cb);
    if (!private_clip) {
	CloseClipboard ();
	return NULL;
    }
    size = strlen (private_clip);
    clip = new char[size+1];
    if (!clip)
	BUG (NULL);
    memset (clip, 0, size+1);
    memcpy (clip, private_clip, size);
    GlobalUnlock(cb);
    CloseClipboard ();
    return clip;
}


/* Set the new clipboard data to @data. */
static void
set_w32_clip_text (const char *data, int size)
{
    HANDLE cb;
    char *private_data;
    
    if (!OpenClipboard(NULL))
	return;        
    cb = GlobalAlloc (GHND, size+1);
    if (!cb)
	BUG (NULL);
    private_data = (char*)GlobalLock (cb);
    if (private_data != NULL) {
	EmptyClipboard ();
	memset (private_data, 0, size+1);
	memcpy (private_data, data, size);
	SetClipboardData (CF_TEXT, cb);
	GlobalUnlock (cb);
    }
    
    CloseClipboard ();
    GlobalFree (cb);
}


/* Search for the armor header 'Version:' in @r_dh and
   add the WinPT version. On success @r_dh is replaced
   with the modified data.
   Return value: 0 on success. */
static gpgme_error_t
gpg_data_change_version (gpgme_data_t *r_dh)
{
    gpgme_error_t err;
    gpgme_data_t mdh;
    const char *s;
    char line[BUFSIZE+256];
    int n;

    if (!r_dh)
	return gpg_error (GPG_ERR_INV_ARG);
    
    err = gpgme_data_new (&mdh);
    if (err)
        return err;

    /* Pattern we search for. */
    s = "Version: GnuPG";
    gpgme_data_rewind (*r_dh);
    while ((n=gpg_data_readline (*r_dh, line, BUFSIZE-1)) > 0) {
        if (strlen (line) > strlen (s) &&
            !strncmp (line, s, strlen (s))  &&
            !stristr (line, "WinPT "PGM_VERSION)) {
	    /* Do not assume a \r\n sequence. */
	    if (strstr (line, "\r\n"))
		line[strlen (line) - 2] = '\0';
	    if (strstr (line, "\n"))
		line[strlen (line) - 1] = '\0';
            strcat (line, " - WinPT "PGM_VERSION"\r\n");
	    // we add a new header to indicate the data is in UTF-8
	    strcat (line, "Charset: UTF-8\r\n");
        }
	gpgme_data_write (mdh, line, strlen (line));
    }

    gpgme_data_write (mdh, "", 1);
    gpgme_data_rewind (mdh);
    gpgme_data_release (*r_dh);
    *r_dh = mdh;
    return err;
}


/* The problem is, that it might be also possible that we
   received an ciphertext which is not encoded with UTF-8.
   Thus, we check explicitly for the 'Charset: UTF-8' header.
   If it is not found, we assume a different encoding and
   no decoding is performed. */
static int 
has_utf8_charset (gpgme_data_t dh)
{    
    char buf[301];
    int found = 0;
    
    for (;;) {
	int n = gpg_data_readline (dh, buf, 300);
	if (n < 1)
	    break;
	// Read until the first blank line, if the
	// header has not been found yet, it does not exist
	if ((n == 1 && buf[0] == '\n') ||
	    (n == 2 && buf[0] == '\r'))
	    break;
	if (!memcmp (buf, "Charset:", 8) && strstr (buf, "UTF-8")) {
	    found = 1;
	    break;
	}
    }
    gpgme_data_rewind (dh);
    return found;
}


/* Retrieve the clipboard data and create a new gpgme
   data object @r_dh and return it. If @wraplen is > 0
   wrap lines of the buffer.
   Return value: 0 on success. */
gpgme_error_t
gpg_data_new_from_clipboard (gpgme_data_t *r_dh, int wraplen)
{
    gpgme_error_t err = 0;
    gpgme_data_t dh;
    char *clip_text = NULL;

    if (!r_dh)
	return gpg_error (GPG_ERR_INV_ARG);
    *r_dh = NULL;
    clip_text = get_w32_clip_text ();
    if (!clip_text)
        return gpg_error (GPG_ERR_NO_DATA);
    err = gpgme_data_new_from_mem (&dh, clip_text, strlen (clip_text), 1);
    if (clip_text)
	free (clip_text);
    if (wraplen > 0)
	gpg_data_wrap_lines (&dh, wraplen);
    
    *r_dh = dh;
    return err;
}

/* Retrieve the clipboard data and create a new gpgme
   data object @r_dh and return it. If @wraplen is > 0
   wrap lines of the buffer.
   Return value: 0 on success. */
gpgme_error_t
gpg_data_utf8_new_from_clipboard (gpgme_data_t *r_dh, 
				  int wraplen, int *r_is_utf8)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    gpgme_data_t dh;
    char *clip_text, *utf8_text;

    if (!r_dh)
	return gpg_error (GPG_ERR_INV_ARG);
    *r_dh = NULL;
    clip_text = get_w32_clip_text ();
    if (!clip_text)
        return gpg_error (GPG_ERR_NO_DATA);
        
    
    /* We automatically encode all clipboard data
       into the UTF-8 format for a portable transportation. */
    utf8_text = native_to_utf8 (clip_text);
    free_if_alloc (clip_text); clip_text = utf8_text;
    
    err = gpgme_data_new_from_mem (&dh, clip_text, strlen (clip_text), 1);
    free_if_alloc (clip_text);
    if (wraplen > 0)
	gpg_data_wrap_lines (&dh, wraplen);
    
    // for ARMORED data find out if the Charset header is set to UTF8
    if (r_is_utf8)
	*r_is_utf8 = has_utf8_charset (dh);
    
    *r_dh = dh;
    return err;
}



gpgme_error_t
gpg_data_release_utf8_to_clipboard (gpgme_data_t dh)
{
    char *clip_text, *utf8_text;
    size_t n;
    
    if (!dh)
	return gpg_error (GPG_ERR_INV_ARG);

    gpgme_data_write (dh, "", 1); // make sure it's a string!
    utf8_text = gpgme_data_release_and_get_mem (dh, &n);
    if (!utf8_text)
	return gpg_error (GPG_ERR_INV_VALUE);
    if (n < 0) {
	gpgme_free (utf8_text);
	return gpg_error (GPG_ERR_NO_DATA);
    }

    /* We assume that all data from GPG is in UTF-8 and
     need to be decoded. If the charset is US-ASCII,
     this wouldn't do any harm, but if the data is ISO-XXX-XX
     we get scrambled output. */
    utf8_text[n] = '\0';
    clip_text = utf8_to_native (utf8_text);
    wipememory (utf8_text, n);
    gpgme_free (utf8_text);
    n = strlen (clip_text);
    set_w32_clip_text (clip_text, n);
    wipememory (clip_text, n);
    free_if_alloc (clip_text);
    return gpg_error (GPG_ERR_NO_ERROR);
}


/* Release a gpgme data object @dh and write the contents of
   the object to the clipboard. If @chg_ver is set, modify the
   Version: header and add WinPT information. */
gpgme_error_t
gpg_data_release_to_clipboard (gpgme_data_t dh, int chg_ver)
{
    gpgme_data_t in;
    char *clip_text;
    size_t n;
    
    if (!dh)
	return gpg_error (GPG_ERR_INV_ARG);
    
    in = dh;
    if (chg_ver)
	gpg_data_change_version (&in);
    
    clip_text = gpgme_data_release_and_get_mem (in, &n);
    if (clip_text && *clip_text) {
        set_w32_clip_text (clip_text, n);
	wipememory (clip_text, n);
	gpgme_free (clip_text);
    }
    
    return gpg_error (GPG_ERR_NO_ERROR);
}
