/* wptKeyPropsDlg.cpp - WinPT key property dialog
 *	Copyright (C) 2000-2003, 2005-2006, 2008 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <assert.h>

#include "resource.h"
#include "wptErrors.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptNLS.h"
#include "wptDlgs.h" 
#include "wptTypes.h"
#include "wptKeylist.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptKeyEdit.h"


/* Context to store associated data of the dialog. */
struct prop_info_s {
    winpt_key_t key;
    char photo_file[MAX_PATH+128+1];
    bool has_photo;
};


/* Check that the key is not expired or revoked. */
static int
do_check_key (gpgme_key_t key)
{
    int invalid;

    invalid = key->expired;
    if (!invalid)
	invalid = key->revoked;
    return invalid;
}


/* Convert a trust integer into a string representation. */
static const char*
ownertrust_to_string (int val, bool is_keypair)
{
    const char *inf;
    int id = val;

    switch (id) {
    case 1: inf = _("Don't know");         break;
    case 2: inf = _("I do NOT trust");     break;
    case 3: inf = _("I trust marginally"); break;
    case 4: inf = _("I trust fully");      break;
    case 5:
    case 6: 
	if (is_keypair)
	    inf = _("I trust ultimately (implicit)");
	else
	    inf = _("I trust ultimately"); break;
    default:inf = _("Unknown");            break;
    }
    
    return inf;
}


int
get_photo_tmpname (gpgme_key_t key, char *buf, size_t buflen)
{
    const char *fmt = "winpt_photo_%p.tmp";
    char name[64];

    if (buflen < (MAX_PATH+strlen(fmt)+8+1))
	return WPTERR_GENERAL;
    _snprintf (name, DIM (name)-1, fmt, key);
    get_temp_name (buf, buflen-1, name);
    return 0;
}




/* Load the photo from the key @key */
int
key_load_photo (winpt_key_t key, 
		char *photo_file, size_t photo_file_size,
		gpgme_validity_t *r_valid)
{
    FILE *fp;
    const BYTE *img;
    DWORD imglen;
    
    img = key->ext->attrib.d;
    imglen = key->ext->attrib.len;
    if (img && !key->ext->attrib.validity)
	get_uat_validity (key->ctx->subkeys->keyid, 
			  &key->ext->attrib.validity);
    if (r_valid)
	*r_valid = key->ext->attrib.validity;

    if (!img || imglen < 1)
	return -1;
            
    get_photo_tmpname (key->ctx, photo_file, photo_file_size);
    fp = fopen (photo_file, "wb");
    if (fp != NULL) {
	const int pos = 16;
	fwrite (img + pos, 1, imglen - pos, fp);
	fclose (fp);
	return 0;
    }
    
    return -1;    
}


/* Return string representation of the key validity. @key. */
static const char*
get_validity (gpgme_key_t key)
{
    if (key->expired)
	return _("Expired");	
    if (key->revoked)
	return _("Revoked");
    if (key->disabled)
	return _("Disabled");
    if (key->invalid)
	return _("Invalid");
    return get_key_trust2 (NULL, key->uids->validity, 0, 0);
}


/* Return true if the key has designated revokers. */
static bool
check_for_desig_rev (gpgme_key_t key)
{
    winpt_key_s k;

    memset (&k, 0, sizeof (k));
    if (!winpt_get_pubkey (key->subkeys->keyid, &k))
	return k.ext->gloflags.has_desig_rev? true : false;
    return false;
}


/* Print information (name) of the smart card. */
static const char*
get_card_type (winpt_key_t k)
{    
    static char buf[64];

    if (!k->ext || !k->ext->card_type)
	return "";
    _snprintf (buf, DIM (buf)-1, _("Card-Type: %s\r\n"), 
	       k->ext->card_type);
    return buf;
}


/* Return 1 if at least one user-ID is valid. */
static int
key_is_valid (gpgme_key_t key)
{
    gpgme_user_id_t u;

    for (u=key->uids; u; u=u->next) {
	if (u->validity >= GPGME_VALIDITY_MARGINAL)
	    return 1;
    }
    return 0;
}


/* Return extended algorithm information. */
const char*
props_get_key_algo (gpgme_key_t key, int idx)
{
    /* PGP calls the old RSAv3 keys 'RSA Legacy' and because this
       is a good method to differ between OpenPGP v4 cert-only keys
       and v3 RSA keys, we use the same notation. */
    if (key->subkeys != NULL && strlen (key->subkeys->fpr) == 32)
	return "RSA Legacy";
    return get_key_algo (key, idx);
}


/* Display the key information for key @k. */
static void
display_key_info (HWND dlg, winpt_key_t k)
{
    gpgme_key_t key;
    struct winpt_key_s sk;
    char info[512];
    const char *inf;
    DWORD created, expires;

    gpg_keycache_update_attr (k->ext, KC_ATTR_PREFSYM, 0);
    memset (&sk, 0, sizeof (sk));
    if (k->key_pair && !winpt_get_seckey (k->keyid, &sk))
	k->is_protected = sk.is_protected;
    key = k->ext->key;
    created = key->subkeys->timestamp;
    expires = key->subkeys->expires;
    _snprintf (info, DIM (info)-1,
               _("Type: %s\r\n"
               "Key ID: 0x%s\r\n"
               "Algorithm: %s\r\n"
               "Size: %s bits\r\n"
               "Created: %s\r\n"
	       "Expires: %s\r\n"
               "Validity: %s\r\n"
	       "%s\r\n"),
               get_key_type (key),
               k->keyid,
               props_get_key_algo (key, 0),
               get_key_size (key, 0),
               get_key_created (created),
	       get_key_expire_date (expires),
               get_validity (key), 
	       get_card_type (&sk));

    SetDlgItemText (dlg, IDC_KEYPROPS_INFO, info);
    SetDlgItemText (dlg, IDC_KEYPROPS_FPR, get_key_fpr (key));  
    inf = ownertrust_to_string (key->owner_trust, k->key_pair);
    SetDlgItemText (dlg, IDC_KEYPROPS_OT, inf);
}



static void
on_init_dialog (HWND dlg, WPARAM wparam, LPARAM lparam)
{
    winpt_key_t k = (winpt_key_t)lparam;
    assert (k != NULL);
    SetWindowText (dlg, _("Key Properties"));
    SetDlgItemText (dlg, IDC_KEYPROPS_OT_CHANGE, _("&Change"));
    SetDlgItemText (dlg, IDC_KEYPROPS_REVOKERS, _("&Revokers"));
    SetDlgItemText (dlg, IDC_KEYPROPS_CHANGE_PWD, _("Change &Password"));
    SetDlgItemText (dlg, IDC_KEYPROPS_OTINF, _("Ownertrust:"));
    SetDlgItemText (dlg, IDC_KEYPROPS_FPRTXT, _("Fingerprint:"));
    
    display_key_info (dlg, k);
    
    if (k->key_pair)
        EnableWindow (GetDlgItem (dlg, IDC_KEYPROPS_CHANGE_PWD), TRUE);
    if (check_for_desig_rev (k->ctx))
	EnableWindow (GetDlgItem (dlg, IDC_KEYPROPS_REVOKERS), TRUE);
    if (do_check_key (k->ctx))
	EnableWindow (GetDlgItem (dlg, IDC_KEYPROPS_OT_CHANGE), FALSE);
    center_window (dlg, NULL);  
    SetForegroundWindow (dlg);
}


/* Dialog box procedure to show the key properties. */
BOOL CALLBACK
keyprops_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    struct prop_info_s *prop = NULL;
    const char *inf;
    gpgme_validity_t valid;
    int rc;
    
    if (msg != WM_INITDIALOG) {
	prop = (struct prop_info_s*)GetWindowLong (dlg, GWL_USERDATA);
	if (prop == NULL)
	    return FALSE;
    }

    switch (msg) {
    case WM_INITDIALOG:
	assert (lparam != 0);
	prop = new struct prop_info_s;
	prop->key = (winpt_key_t)lparam;
	prop->has_photo = 0;
        SetWindowLong (dlg, GWL_USERDATA, (LONG)prop);
        on_init_dialog (dlg, wparam, lparam);
	if (!key_load_photo (prop->key, 
			     prop->photo_file, DIM (prop->photo_file)-1,
			     &valid)) {
	    prop->key->has_photo = 1;
	    prop->has_photo = true;
	    if (valid < GPGME_VALIDITY_MARGINAL)
		SetDlgItemText (dlg, IDC_KEYPROPS_IMGINF, 
				_("Photo-ID not validated."));
	}
        return TRUE;

    case WM_DESTROY:
	if (prop->has_photo)
	    DeleteFile(prop->photo_file);
	delete prop;prop = NULL;
	SetWindowLong (dlg, GWL_USERDATA, 0);
	break;
      
    case WM_PAINT:
	/* Display the photo in the frame of the dialog @dlg.
	   The coordinates are fixed to (0,0). */
	if (prop->has_photo) {
	    POINT p;
	    p.x = p.y = 0;
	    /* In case of errors we disable the flag to
	       avoid an infinite loop. */
	    if (jpg_show (GetDlgItem (dlg, IDC_KEYPROPS_IMG), 
			  &p, prop->photo_file))
		prop->has_photo = false;
	}
	break;

    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDOK:
            EndDialog (dlg, TRUE);
            return TRUE;

	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    return TRUE;
            
        case IDC_KEYPROPS_OT_CHANGE:
	    if (!prop->key->key_pair && !key_is_valid (prop->key->ctx)) {
		rc = msg_box (dlg, _("This is a non-valid key.\n"
				     "Modifying the ownertrust has no effect on such keys.\n\n"
			             "Do you really want to continue?"), 
				     _("WinPT Warning"), MB_ICONWARNING|MB_YESNO);
		if (rc == IDNO)
		    return TRUE;
	    }
            rc = dialog_box_param (glob_hinst, 
				   (LPCSTR)IDD_WINPT_KEYEDIT_OWNERTRUST,
				    dlg, keyedit_ownertrust_dlg_proc,
				   (LPARAM)prop->key, _("Change Ownertrust"),
				    IDS_WINPT_KEYEDIT_OWNERTRUST);
	    if (rc == FALSE) /* Cancel */
		return TRUE;

	    inf = ownertrust_to_string (prop->key->callback.new_val, 
					prop->key->key_pair);
            SetDlgItemText (dlg, IDC_KEYPROPS_OT, inf);
            msg_box (dlg, _("Ownertrust successfully changed."), 
		     _("GnuPG Status"), MB_OK);
	    prop->key->update = 1;
            return TRUE;
            
        case IDC_KEYPROPS_CHANGE_PWD:
            keyedit_change_passwd (prop->key, dlg);	    
            return TRUE;

	case IDC_KEYPROPS_REVOKERS:
	    prop->key->update = dialog_box_param (glob_hinst, 
					  (LPCTSTR)IDD_WINPT_KEYREVOKERS, dlg,
					  key_revokers_dlg_proc, (LPARAM)prop->key,
					  _("Key Revokers"), 
					  IDS_WINPT_KEY_REVOKERS);
	    UpdateWindow (dlg);
	    break;
        }
    }
    
    return FALSE;
}
