/* wptKeyRevokersDlg.cpp - Designated Revoker Keys
 *	Copyright (C) 2002, 2003, 2005, 2007 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptCommonCtl.h"
#include "wptNLS.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptErrors.h"
#include "wptKeyserver.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptKeyEdit.h"
#include "wptKeylist.h"
#include "wptKeyManager.h"

/* symbolic column IDs */
enum revoke_col_t {
    REV_COL_NAME    = 0,
    REV_COL_KEYID   = 1,
    REV_COL_ALGO    = 2
};

/* Initialize the listview given in @ctrl and return it in @r_lv. */   
static void
revokelist_build (listview_ctrl_t *r_lv, HWND ctrl)
{
    listview_ctrl_t lv;
    struct listview_column_s rlist[] = {
	{0, 160, (char *)_("Name") },
	{1,  80, (char *)_("Key ID")},
	{2,  55, (char *)_("Algorithm")},
	{0}
    };
    int i;

    listview_new (&lv, ctrl);
    lv->items = DIM (rlist)-1;
    for (i = 0; i < lv->items; i++)
	listview_add_column (lv, &rlist[i]);
    listview_set_ext_style (lv);
    *r_lv = lv;
}


/* Get the key revokers of the key @key.
   If reload is 1, an update is forced otherwise the
   cached revokers are returned. */
gpgme_error_t
key_get_revokers (winpt_key_t key, int reload, gpg_desig_rev_t *r_rev)
{
    gpgme_error_t err;
    gpg_desig_rev_t rev;
    GpgKeyEdit ke;

    if (!reload && key->ext->rev) {
	*r_rev = key->ext->rev;
	return 0;
    }
    ke.setKeyID (key->keyid);
    err = ke.getDesignatedRevoker (&rev);
    if (err) {
	gpg_desig_rev_release (rev);
	return err;
    }
    if (key->ext->rev)
	gpg_desig_rev_release (rev);
    key->ext->rev = rev;
    *r_rev = rev;
    return 0;
}


/* Initialize the listview @ctrl with all designated revokers of the
   key @key.
   Return value: the initialized listview or NULL on error. */
static listview_ctrl_t
revokelist_init (HWND ctrl, winpt_key_t key)
{    
    gpgme_error_t err;
    struct winpt_key_s revkey;
    gpg_desig_rev_t rev=NULL, r;
    listview_ctrl_t lv;
    const char *alg;
    const char *uid = NULL, *kid;
    char keyid[32];

    err = key_get_revokers (key, 0, &rev);
    if (err) {
	msg_box (NULL, gpgme_strerror (err), 
		 _("Designated Key Revokers"), MB_ERR);
	return NULL;
    }
    revokelist_build (&lv, ctrl);
    for (r = rev; r; r = r->next) {
	memset (&revkey, 0, sizeof (revkey));
	kid = get_keyid_from_fpr (r->fpr);
	if (winpt_get_pubkey (kid, &revkey))
	    uid = _("user ID not found");
	else
	    uid = revkey.ext->uids->name;
	_snprintf (keyid, DIM (keyid)-1, "0x%s", kid);
	listview_add_item2 (lv, "", revkey.ext);
	alg = get_key_pubalgo (r->pubkey_algo);
	listview_add_sub_item (lv, 0, REV_COL_ALGO, alg);
	listview_add_sub_item (lv, 0, REV_COL_KEYID, keyid);
	listview_add_sub_item (lv, 0, REV_COL_NAME, uid);
    }
    return lv;
}


/* Dialog box procedure to show designated revokers. */
BOOL CALLBACK
key_revokers_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv;
    static int key_update = 0;
    winpt_key_t key;
    char buf[128], keyid[32];
    int rc;
    
    switch (msg) {
    case WM_INITDIALOG:	
	key = (winpt_key_t)lparam;
	if (!key)
	    BUG (0);
	lv = revokelist_init (GetDlgItem (dlg, IDC_KEYREVOKERS_LIST), key);
	SetDlgItemText (dlg, IDC_KEYREVOKERS_INF, _("Designated Revoker Keys"));
	SetWindowText (dlg, _("Key Revokers"));
	SetForegroundWindow (dlg);
	return TRUE;

    case WM_NOTIFY:
	NMHDR *notify;
	notify = (NMHDR *)lparam;
	if (notify && notify->code == (UINT)NM_DBLCLK && 
	    notify->idFrom == IDC_KEYREVOKERS_LIST) {
	    int idx = listview_get_curr_pos (lv);

	    listview_get_item_text (lv, idx, REV_COL_KEYID, keyid, DIM (keyid)-1);
	    listview_get_item_text (lv, idx, REV_COL_NAME, buf, DIM (buf)-1);
	    if (!strcmp (buf, _("user ID not found"))) {
		int id = log_box (_("Designated Key Revokers"), MB_YESNO|MB_INFO,
				  _("Do you want to retrieve %s via the default keyserver?"), keyid);
		if (id == IDNO)
		    break;
		rc = hkp_recv_key (dlg, default_keyserver,
				   default_keyserver_port, keyid, 0, 0);
		if (!rc) {
		    struct winpt_key_s pk;

		    keycache_update (0, keyid);
		    key_update = 1;
		    memset (&pk, 0, sizeof (pk));
		    if (winpt_get_pubkey (keyid, &pk))
			BUG (NULL);
		    /* XXX: utf8 decode problem. */
		    if (pk.ext && pk.ext->uids)
			listview_add_sub_item (lv, idx, REV_COL_NAME, pk.ext->uids->name);
		}
	    }
	    else {
		struct winpt_key_s k;

		memset (&k, 0, sizeof k);
		km_get_key (lv, idx, &k);
		DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_KEYPROPS, dlg,
				keyprops_dlg_proc, (LPARAM)&k);
	    }
	}
	break;

    case WM_DESTROY:
	if (lv) {
	    listview_release (lv);
	    lv = NULL;
	}
	return FALSE;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    EndDialog (dlg, key_update);
	    break;

	case IDCANCEL:
	    EndDialog (dlg, 0);
	    break;
	}
	break;
    }

    return FALSE;
}
